import * as i0 from '@angular/core';
import { inject, Injectable, PLATFORM_ID, NgModule, Inject, makeEnvironmentProviders, ENVIRONMENT_INITIALIZER } from '@angular/core';
import { Router, ChildActivationEnd } from '@angular/router';
import { filter, map, take } from 'rxjs/operators';
import * as fs from 'fs';
import { join } from 'path';
import { APP_BASE_HREF, isPlatformServer, DOCUMENT } from '@angular/common';
import { ɵSERVER_CONTEXT, BEFORE_APP_SERIALIZED } from '@angular/platform-server';
import { HTTP_INTERCEPTORS } from '@angular/common/http';
import { catchError, throwError } from 'rxjs';

const initialState = {
    revalidate: null,
    errors: [],
    extra: {},
};
class NgxIsrService {
    constructor() {
        this.router = inject(Router);
        this.state = initialState;
    }
    getState() {
        return this.state;
    }
    patchState(partialState) {
        this.state = { ...this.state, ...partialState };
    }
    getExtra() {
        return this.state.extra;
    }
    /**
     * Activate the service and listen to router events
     * @returns void
     */
    activate() {
        this.router.events
            .pipe(filter((e) => e instanceof ChildActivationEnd), map((event) => {
            let snapshot = event.snapshot;
            // get the last child route
            while (snapshot.firstChild !== null) {
                snapshot = snapshot.firstChild;
            }
            // get the data from the last child route
            return snapshot.data;
        }), take(1))
            .subscribe((data) => {
            // if revalidate is defined, set it
            if (data?.['revalidate'] !== undefined) {
                this.patchState({ revalidate: data['revalidate'] });
            }
        });
    }
    /**
     * Add error to the state
     * @param error HttpErrorResponse
     * @returns void
     * @example
     * ```typescript
     * this.isrService.addError(err);
     * ```
     */
    addError(error) {
        this.patchState({ errors: [...this.getState().errors, error] });
    }
    /**
     * Add extra data to the state
     * @param extra Record<string, any>
     * @returns void
     * @example
     * ```typescript
     * this.isrService.addExtra({ foo: 'bar' });
     * ```
     */
    addExtra(extra = {}) {
        this.patchState({ extra: { ...this.getExtra(), ...extra } });
    }
}
NgxIsrService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrService, deps: [], target: i0.ɵɵFactoryTarget.Injectable });
NgxIsrService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }] });

class CacheHandler {
}

/**
 * Get ISR options from html string
 * @param html
 * @returns
 * @example
 * ```typescript
 * const { revalidate, errors } = getISROptions(html);
 * ```
 * @internal
 */
function getRouteISRDataFromHTML(html) {
    const indexOfScriptTag = html?.indexOf(ISR_SCRIPT_TAG);
    // check if script tag is not included
    if (!html || indexOfScriptTag === -1) {
        return { revalidate: null, errors: [] };
    }
    // start from script till the end of html file
    const isrScript = html.substring(indexOfScriptTag);
    // first occurrence of closing script tag
    const indexOfCloseScriptTag = isrScript.indexOf('</script>');
    const val = isrScript
        .substring(0, indexOfCloseScriptTag) // remove close script tag
        .replace(ISR_SCRIPT_TAG, ''); // remove start script tag
    return JSON.parse(val);
}
/**
 * Script tag that will be included in the page if one of the routes on the page
 * has `revalidate` key in its route data
 *
 * @internal
 */
const ISR_SCRIPT_TAG = '<script id="isr-state" type="application/json">';

class FileSystemCacheHandler extends CacheHandler {
    constructor(options) {
        super();
        this.options = options;
        this.cache = new Map();
        if (!options.cacheFolderPath) {
            throw new Error('Cache folder path is required!');
        }
        if (options.addPrerenderedPagesToCache && !options.prerenderedPagesPath) {
            throw new Error('Prerendered pages path is required when `addPrerenderedPagesToCache` is enabled!');
        }
        this.options = options;
        this.populateCacheFromFilesystem();
    }
    get cacheFolderPath() {
        return this.options.cacheFolderPath;
    }
    async add(route, html, config) {
        return new Promise((resolve, reject) => {
            route = route.charAt(0) === '/' ? route.slice(1) : route;
            const fileName = convertRouteToFileName(route);
            const filePath = `${this.getFileFullPath(fileName)}.html`;
            fs.writeFile(filePath, html, 'utf-8', (err) => {
                if (err) {
                    reject('Error: 💥 The request was not cached!');
                }
                this.cache.set(route, {
                    html: filePath,
                    options: config || { revalidate: null },
                    createdAt: Date.now(),
                });
                console.log('The request was cached!', route);
                resolve();
            });
        });
    }
    get(route) {
        return new Promise(async (resolve, reject) => {
            const cachedUrl = this.cache.get(route);
            if (cachedUrl) {
                // on html field we have saved path to file
                const html = await this.readFromFile(cachedUrl.html);
                const cacheData = {
                    ...cachedUrl,
                    html,
                };
                resolve(cacheData);
            }
            else {
                reject('Error: 💥 Url is not cached.');
            }
        });
    }
    has(url) {
        return Promise.resolve(this.cache.has(url));
    }
    delete(url) {
        return new Promise((resolve, reject) => {
            const cachedUrl = this.cache.get(url);
            if (cachedUrl) {
                const filePath = this.getFileFullPath(url);
                fs.unlink(cachedUrl.html, (err) => {
                    if (err) {
                        reject('Error: 💥 Cannot delete file' + filePath);
                    }
                    else {
                        this.cache.delete(url);
                        resolve(true);
                    }
                });
            }
            else {
                reject('Error: 💥 Url is not cached.');
            }
        });
    }
    getAll() {
        const data = Array.from(this.cache.keys());
        return Promise.resolve(data);
    }
    populateCacheFromFilesystem() {
        if (!fs.existsSync(this.cacheFolderPath)) {
            console.log('Cache folder does not exist. Creating...');
            fs.mkdir(this.cacheFolderPath, (err) => {
                if (!err) {
                    console.log('Cache folder was created.');
                }
            });
        }
        if (this.options.addPrerenderedPagesToCache) {
            console.log('Adding prerendered pages to cache...');
            this.addPrerenderedPagesToCache();
        }
    }
    clearCache() {
        throw new Error('Method not implemented.');
    }
    addPrerenderedPagesToCache() {
        // read all folders in browser folder and check if they have index.html inside
        // if yes add the folder name to cache as url and index.html as html
        // then remove the found files because they will be handled by ISR
        const folderPath = this.options.prerenderedPagesPath;
        const pathsToCache = [];
        try {
            const files = fs.readdirSync(folderPath);
            for (const file of files) {
                const filePath = join(folderPath, file);
                const isDirectory = fs.statSync(filePath).isDirectory();
                if (isDirectory) {
                    const indexHtmlFiles = findIndexHtmlFilesRecursively(filePath);
                    pathsToCache.push(...indexHtmlFiles.map(({ path, html }) => ({
                        path: path.replace(folderPath, ''),
                        html,
                    })));
                }
            }
        }
        catch (err) {
            console.error('ERROR! 💥 ! Cannot read folder: ' + folderPath);
        }
        for (const { path, html } of pathsToCache) {
            const routePath = path.substring(0).replace('/index.html', '');
            const { revalidate, errors } = getRouteISRDataFromHTML(html);
            this.add(routePath, html, { revalidate, errors });
        }
        for (const { path } of pathsToCache) {
            const file = join(this.options.prerenderedPagesPath, path);
            fs.rmSync(file, { recursive: true, force: true });
        }
    }
    async readFromFile(filePath) {
        return new Promise((resolve, reject) => {
            fs.readFile(filePath, 'utf-8', (err, data) => {
                if (err) {
                    console.error('ERROR! 💥 ! Cannot read file: ' + filePath);
                    reject(err);
                }
                resolve(data);
            });
        });
    }
    getFileFullPath(fileName) {
        // fileName may have a / in the beginning, so we want to remove it first
        if (fileName.charAt(0) === '/') {
            fileName = fileName.slice(1);
        }
        return join(this.cacheFolderPath, '/', fileName);
    }
}
function findIndexHtmlFilesRecursively(path) {
    const data = [];
    try {
        const files = fs.readdirSync(path);
        files.forEach((file) => {
            const filePath = join(path, file);
            if (fs.statSync(filePath).isDirectory()) {
                data.push(...findIndexHtmlFilesRecursively(filePath));
            }
            else if (file.includes('index.html')) {
                const html = fs.readFileSync(filePath, 'utf8');
                data.push({ path: filePath, html });
            }
        });
    }
    catch (err) {
        console.error('ERROR! 💥 ! Cannot read folder: ' + path);
        return [];
    }
    return data;
}
function convertRouteToFileName(route) {
    return route.replace(new RegExp('/', 'g'), '__');
}

const defaultCacheISRConfig = {
    revalidate: null,
    buildId: null,
};
class InMemoryCacheHandler {
    constructor() {
        this.cache = new Map();
    }
    add(url, html, config = defaultCacheISRConfig) {
        const htmlWithMsg = html + cacheMsg(config.revalidate);
        return new Promise((resolve, reject) => {
            const cacheData = {
                html: htmlWithMsg,
                options: config,
                createdAt: Date.now(),
            };
            this.cache.set(url, cacheData);
            resolve();
        });
    }
    get(url) {
        return new Promise((resolve, reject) => {
            if (this.cache.has(url)) {
                resolve(this.cache.get(url));
            }
            reject('This url does not exist in cache!');
        });
    }
    getAll() {
        return new Promise((resolve, reject) => {
            resolve(Array.from(this.cache.keys()));
        });
    }
    has(url) {
        return new Promise((resolve, reject) => {
            resolve(this.cache.has(url));
        });
    }
    delete(url) {
        return new Promise((resolve, reject) => {
            resolve(this.cache.delete(url));
        });
    }
}
const cacheMsg = (revalidateTime) => {
    const time = new Date().toISOString().replace(/T/, ' ').replace(/\..+/, '');
    let msg = '<!-- ';
    msg += `\n🚀 NgxISR: Served from cache! \n⌛ Last updated: ${time}. `;
    if (revalidateTime)
        msg += `\n⏭️ Next refresh is after ${revalidateTime} seconds. `;
    msg += ' \n-->';
    return msg;
};

const EXTRA_PROVIDERS = [
    { provide: ɵSERVER_CONTEXT, useValue: 'ngx-isr' },
];
// helper method that generates html of an url
const renderUrl = async (options) => {
    const { req, res, url, indexHtml, providers } = options;
    // we need to override url of req with the one we have in parameters
    req.url = url;
    req.originalUrl = url;
    const BASE_URL_PROVIDER = {
        provide: APP_BASE_HREF,
        useValue: req.baseUrl,
    };
    return new Promise((resolve, reject) => {
        const allProviders = providers
            ? [...providers, ...EXTRA_PROVIDERS] // if providers are provided, we add them to the list
            : [...EXTRA_PROVIDERS, BASE_URL_PROVIDER]; // if not, we add the default providers
        res.render(indexHtml, { req, providers: allProviders }, async (err, html) => {
            if (err) {
                reject(err);
            }
            resolve(html);
        });
    });
};

class CacheRegeneration {
    constructor(isrConfig, cache, indexHtml) {
        this.isrConfig = isrConfig;
        this.cache = cache;
        this.indexHtml = indexHtml;
        // TODO: make this pluggable because on serverless environments we can't share memory between functions
        // so we need to use a database or redis cache to store the urls that are on hold if we want to use this feature
        this.urlsOnHold = []; // urls that have regeneration loading
    }
    async regenerate(req, res, cacheData, logger, providers) {
        const { url } = req;
        if (this.urlsOnHold.includes(url)) {
            logger.log('Another regeneration is on-going for this url...');
            return;
        }
        const { revalidate } = cacheData.options;
        logger.log(`The url: ${url} is being regenerated.`);
        this.urlsOnHold.push(url);
        renderUrl({ req, res, url, indexHtml: this.indexHtml, providers }).then((html) => {
            const { errors } = getRouteISRDataFromHTML(html);
            // if there are errors, don't add the page to cache
            if (errors?.length) {
                // remove url from urlsOnHold because we want to try to regenerate it again
                this.urlsOnHold = this.urlsOnHold.filter((x) => x !== url);
                logger.log('💥 ERROR: Url: ' + url + ' was not regenerated!', errors);
                return;
            }
            // add the regenerated page to cache
            this.cache
                .add(req.url, html, { revalidate, buildId: this.isrConfig.buildId })
                .then(() => {
                // remove from urlsOnHold because we are done
                this.urlsOnHold = this.urlsOnHold.filter((x) => x !== url);
                logger.log('Url: ' + url + ' was regenerated!');
            });
        });
    }
}

/**
 * Logger class for ISR package
 * @internal
 */
class ISRLogger {
    constructor(showLogs) {
        this.showLogs = showLogs;
    }
    /**
     * Log a message to the console
     * @param message The message to log
     * @param optionalParams Optional parameters to log
     * @internal
     */
    log(message, ...optionalParams) {
        this.showLogs && console.log(message, ...optionalParams);
    }
}

class ISRHandler {
    constructor(config) {
        this.config = config;
        this.logger = new ISRLogger(this.config?.enableLogging || false);
        if (!config) {
            throw new Error('Provide ISRHandlerConfig!');
        }
        // if skipCachingOnHttpError is not provided it will default to true
        config.skipCachingOnHttpError = config.skipCachingOnHttpError !== false;
        // if buildId is not provided it will default to null
        config.buildId = config.buildId || null;
        // if invalidateSecretToken is not provided it will default to null
        config.invalidateSecretToken = config.invalidateSecretToken || null;
        if (config.cache && config.cache instanceof CacheHandler) {
            this.logger.log('Using custom cache handler!');
            this.cache = config.cache;
        }
        else {
            this.logger.log('Using in memory cache handler!');
            this.cache = new InMemoryCacheHandler();
        }
        this.cacheRegeneration = new CacheRegeneration(this.config, this.cache, config.indexHtml);
    }
    async invalidate(req, res, config) {
        const { token, urlsToInvalidate } = extractDataFromBody(req);
        const { indexHtml } = this.config;
        if (token !== this.config.invalidateSecretToken) {
            return res.json({
                status: 'error',
                message: 'Your secret token is wrong!!!',
            });
        }
        if (!urlsToInvalidate || !urlsToInvalidate.length) {
            return res.json({
                status: 'error',
                message: 'Please add `urlsToInvalidate` in the payload!',
            });
        }
        const notInCache = [];
        const urlWithErrors = {};
        for (const url of urlsToInvalidate) {
            // check if the url is in cache
            const urlExists = await this.cache.has(url);
            if (!urlExists) {
                notInCache.push(url);
                continue;
            }
            try {
                // re-render the page again
                const html = await renderUrl({
                    req,
                    res,
                    url,
                    indexHtml,
                    providers: config?.providers,
                });
                // get revalidate data in order to set it to cache data
                const { revalidate, errors } = getRouteISRDataFromHTML(html);
                // if there are errors when rendering the site we throw an error
                if (errors?.length && this.config.skipCachingOnHttpError) {
                    urlWithErrors[url] = errors;
                }
                // add the regenerated page to cache
                const cacheConfig = {
                    revalidate,
                    buildId: this.config.buildId,
                };
                await this.cache.add(req.url, html, cacheConfig);
            }
            catch (err) {
                urlWithErrors[url] = err;
            }
        }
        const invalidatedUrls = urlsToInvalidate.filter((url) => !notInCache.includes(url) && !urlWithErrors[url]);
        if (notInCache.length) {
            this.logger.log(`Urls: ${notInCache.join(', ')} does not exist in cache.`);
        }
        if (Object.keys(urlWithErrors).length) {
            this.logger.log(`Urls: ${Object.keys(urlWithErrors).join(', ')} had errors while regenerating!`);
        }
        if (invalidatedUrls.length) {
            this.logger.log(`Urls: ${invalidatedUrls.join(', ')} were regenerated!`);
        }
        const response = {
            status: 'success',
            notInCache,
            urlWithErrors,
            invalidatedUrls,
        };
        return res.json(response);
    }
    async serveFromCache(req, res, next, config) {
        try {
            const cacheData = await this.cache.get(req.url);
            const { html, options, createdAt } = cacheData;
            if (options.buildId !== null && options.buildId !== this.config.buildId) {
                // Cache is from a different build. Serve user using SSR
                next();
                // TODO: We can extend this to serve a page from old build if we want to support that
                return;
            }
            // if the cache is expired, we will regenerate it
            if (options.revalidate && options.revalidate > 0) {
                const lastCacheDateDiff = (Date.now() - createdAt) / 1000; // in seconds
                if (lastCacheDateDiff > options.revalidate) {
                    await this.cacheRegeneration.regenerate(req, res, cacheData, this.logger, config?.providers);
                }
            }
            // Apply the callback if given
            let finalHtml = html;
            if (config?.modifyCachedHtml) {
                const timeStart = performance.now();
                finalHtml = config.modifyCachedHtml(req, html);
                const totalTime = (performance.now() - timeStart).toFixed(2);
                finalHtml += `<!--\nℹ️ NgxISR: This cachedHtml has been modified with modifyCachedHtml()\n❗️ 
        This resulted into more ${totalTime}ms of processing time.\n-->`;
            }
            // Cache exists. Send it.
            this.logger.log(`Page was retrieved from cache: `, req.url);
            return res.send(finalHtml);
        }
        catch (error) {
            // Cache does not exist. Serve user using SSR
            next();
        }
    }
    async render(req, res, next, config) {
        const renderUrlConfig = {
            req,
            res,
            url: req.url,
            indexHtml: this.config.indexHtml,
            providers: config?.providers,
        };
        renderUrl(renderUrlConfig).then(async (html) => {
            const { revalidate, errors } = getRouteISRDataFromHTML(html);
            // Apply the callback if given
            const finalHtml = config?.modifyGeneratedHtml
                ? config.modifyGeneratedHtml(req, html)
                : html;
            // if we have any http errors when rendering the site, and we have skipCachingOnHttpError enabled
            // we don't want to cache it, and, we will fall back to client side rendering
            if (errors?.length && this.config.skipCachingOnHttpError) {
                this.logger.log('Http errors: \n', errors);
                return res.send(finalHtml);
            }
            // if revalidate is null we won't cache it
            // if revalidate is 0, we will never clear the cache automatically
            // if revalidate is x, we will clear cache every x seconds (after the last request) for that url
            if (revalidate === null || revalidate === undefined) {
                // don't do !revalidate because it will also catch "0"
                return res.send(finalHtml);
            }
            // Cache the rendered `html` for this request url to use for subsequent requests
            await this.cache.add(req.url, finalHtml, {
                revalidate,
                buildId: this.config.buildId,
            });
            return res.send(finalHtml);
        });
    }
}
const extractDataFromBody = (req) => {
    const { urlsToInvalidate, token } = req.body;
    return { urlsToInvalidate, token };
};

class HttpErrorsInterceptor {
    constructor(ngxIsrService) {
        this.ngxIsrService = ngxIsrService;
    }
    intercept(request, next) {
        return next.handle(request).pipe(catchError(err => {
            this.ngxIsrService.addError(err);
            return throwError(() => err);
        }));
    }
}
HttpErrorsInterceptor.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: HttpErrorsInterceptor, deps: [{ token: NgxIsrService }], target: i0.ɵɵFactoryTarget.Injectable });
HttpErrorsInterceptor.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: HttpErrorsInterceptor });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: HttpErrorsInterceptor, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: NgxIsrService }]; } });
const HTTP_ERROR_PROVIDER_ISR = {
    provide: HTTP_INTERCEPTORS,
    useClass: HttpErrorsInterceptor,
    multi: true,
};

function addIsrDataBeforeSerialized(isrService, doc) {
    return () => addISRDataToBody(doc, isrService.getState());
}
// append script with revalidate and errors data for the current route
function addISRDataToBody(doc, { revalidate, errors, extra }) {
    return new Promise(resolve => {
        const script = doc.createElement('script');
        script.id = 'isr-state';
        script.setAttribute('type', 'application/json');
        let toBeSerialized = { revalidate };
        if (errors.length)
            toBeSerialized = { ...toBeSerialized, errors };
        if (Object.keys(extra).length)
            toBeSerialized = { ...toBeSerialized, extra };
        script.textContent = JSON.stringify(toBeSerialized);
        doc.body.appendChild(script);
        resolve();
    });
}

class NgxIsrModule {
    constructor(isrService, platformId) {
        // Activate ISR only on the server
        if (isPlatformServer(platformId)) {
            isrService.activate();
        }
    }
    static forRoot() {
        return {
            ngModule: NgxIsrModule,
            providers: [
                NgxIsrService,
                HTTP_ERROR_PROVIDER_ISR,
                {
                    provide: BEFORE_APP_SERIALIZED,
                    useFactory: addIsrDataBeforeSerialized,
                    multi: true,
                    deps: [NgxIsrService, DOCUMENT],
                },
            ],
        };
    }
}
NgxIsrModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrModule, deps: [{ token: NgxIsrService }, { token: PLATFORM_ID }], target: i0.ɵɵFactoryTarget.NgModule });
NgxIsrModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrModule });
NgxIsrModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrModule, providers: [NgxIsrService] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: NgxIsrModule, decorators: [{
            type: NgModule,
            args: [{ providers: [NgxIsrService] }]
        }], ctorParameters: function () { return [{ type: NgxIsrService }, { type: undefined, decorators: [{
                    type: Inject,
                    args: [PLATFORM_ID]
                }] }]; } });

/**
 * @description
 * This function registers the providers needed for ISR to work.
 *
 * @returns {EnvironmentProviders} The providers for the application.
 *
 * @example
 * ```
 * import { provideISR } from 'ngx-isr';
 *
 * @NgModule({
 *  providers: [ provideISR() ]
 * })
 * export class AppServerModule {}
 * ```
 *
 * To configure ISR in a standalone application:
 * ```
 * const serverConfig: ApplicationConfig = {
 *   providers: [
 *     provideServerRendering()
 *     provideISR()
 *   ],
 * };
 * export const config = mergeApplicationConfig(appConfig, serverConfig);
 * ```
 */
const provideISR = () => {
    return makeEnvironmentProviders([
        NgxIsrService,
        HTTP_ERROR_PROVIDER_ISR,
        {
            provide: BEFORE_APP_SERIALIZED,
            useFactory: addIsrDataBeforeSerialized,
            multi: true,
            deps: [NgxIsrService, DOCUMENT],
        },
        {
            provide: ENVIRONMENT_INITIALIZER,
            useFactory: (isrService, platformId) => () => {
                // Activate ISR only on the server
                if (isPlatformServer(platformId)) {
                    isrService.activate();
                }
            },
            deps: [NgxIsrService, PLATFORM_ID],
            multi: true,
        },
    ]);
};

/*
 * Public API Surface of ngx-isr
 */

/**
 * Generated bundle index. Do not edit.
 */

export { CacheHandler, FileSystemCacheHandler, ISRHandler, InMemoryCacheHandler, NgxIsrModule, NgxIsrService, provideISR };
//# sourceMappingURL=ngx-isr.mjs.map
